MODULE Terminal;

IMPORT G := Graph, Out (*!FIXME remove Out*);

CONST
  charW* = 8; charH* = 16;
  cursorTicks* = 30; (* Ticks before cursor flashes *)

TYPE
  ScreenChar* = RECORD
    ch*: CHAR;
    fg*, bg*: INTEGER (* Цвет текста и цвет фона*)
  END;
  ScreenChars* = POINTER TO ARRAY OF ARRAY OF ScreenChar;

VAR
  screen*: G.Bitmap;
  charsX*, charsY*: INTEGER;
  scrH*: INTEGER; (* Screen Height (temporary charsY - 4) *)
  font: G.Font;

  chars: ScreenChars;
  cursorX, cursorY: INTEGER;
  cursorShown: BOOLEAN;
  insertCursor: BOOLEAN;
  cursorTick*: INTEGER; (* To show and hide cursor all the time *)
  mouseX*, mouseY*: INTEGER; (* In chars *)

PROCEDURE ExpandColor*(color: INTEGER): INTEGER;
VAR r, g, b: INTEGER;
BEGIN
  CASE color OF
     0: b :=   0; g :=   0; r :=   0;
  |  1: b :=  90; g :=   0; r :=   0;
  |  2: b :=   0; g := 176; r :=   0;
  |  3: b := 176; g := 176; r :=   0;
  |  4: b :=   0; g :=   0; r := 176;
  |  5: b := 176; g :=   0; r := 176;
  |  6: b :=   0; g :=  85; r := 176;
  |  7: b := 176; g := 176; r := 176;
  |  8: b :=  85; g :=  85; r :=  85;
  |  9: b := 255; g :=  85; r :=  85;
  | 10: b :=  85; g := 255; r :=  85;
  | 11: b := 255; g := 255; r :=  85;
  | 12: b :=  85; g :=  85; r := 255;
  | 13: b := 255; g :=  85; r := 255;
  | 14: b :=  85; g := 255; r := 255;
  | 15: b := 255; g := 255; r := 255;
  ELSE  b :=  85; g :=   0; r := 255;
  END;
  RETURN G.MakeCol(r, g, b)
END ExpandColor;

PROCEDURE InvertColor*(color: INTEGER): INTEGER;
BEGIN
  ASSERT((color >= 0) & (color < 16), 59);
  IF color >= 8 THEN color := color - 8 END; (* Darken *)
  color := 7 - color; (* Invert *)
  RETURN color
END InvertColor;

PROCEDURE DrawChar*(ch: CHAR; x, y, fg, bg: INTEGER);
VAR color: INTEGER;
BEGIN
  G.RectFill(screen, x, y, x + charW - 1,
    y + charH - 1, ExpandColor(bg));
  IF ch # ' ' THEN
    G.DrawCharacter(screen, font, x, y, ch, ExpandColor(fg))
  END
END DrawChar;

PROCEDURE DrawMouse;
VAR x, y, bg, fg: INTEGER; ch: CHAR;
BEGIN
  IF (mouseX >= 0) & (mouseX < charsX) &
     (mouseY >= 0) & (mouseY < charsY) THEN
    bg := InvertColor(chars[mouseY, mouseX].bg);
    fg := InvertColor(chars[mouseY, mouseX].fg);
    ch := chars[mouseY, mouseX].ch
  ELSE bg := 6; fg := 0; ch := ' ' END;
  x := mouseX * charW;  y := mouseY * charH;
  DrawChar(ch, x, y, fg, bg)
END DrawMouse;

PROCEDURE Draw*;
VAR x, y, color: INTEGER;
BEGIN
  (* Chars *)
  FOR y := 0 TO charsY - 1 DO
    FOR x := 0 TO charsX - 1 DO
      DrawChar(chars[y, x]. ch, x * charW, y * charH,
        chars[y, x].fg, chars[y, x].bg)
    END
  END;

  (* Text Cursor *)
  IF cursorShown THEN
    color := ExpandColor(chars[cursorY, cursorX].fg);
    x := cursorX * charW;  y := cursorY * charH;
    IF insertCursor THEN
      G.RectFill(screen, x, y, x + charW - 1, y + charH - 1, color)
    ELSE
      INC(y, charH);
      G.RectFill(screen, x, y - 2, x + charW - 1, y - 1, color)
    END
  END;

  (* Mouse Cursor *)
  DrawMouse
END Draw;

PROCEDURE Act*;
BEGIN
  IF cursorTick >= cursorTicks THEN
    cursorTick := 0; cursorShown := ~cursorShown
  ELSE INC(cursorTick) END
END Act;

PROCEDURE ResetCursorBlink*;
BEGIN
  cursorTick := 0; cursorShown := TRUE
END ResetCursorBlink;

PROCEDURE GoToXY*(x, y: INTEGER);
BEGIN
  ASSERT((x >= 0) & (x < charsX) & (y >= 0) & (y < charsY), 51);
  cursorX := x;  cursorY := y
END GoToXY;

PROCEDURE MouseXY*(x, y: INTEGER);
BEGIN
  mouseX := x;  mouseY := y
END MouseXY;

PROCEDURE ResizeScreen;
BEGIN
  NEW(chars, charsY, charsX)
END ResizeScreen;

PROCEDURE SetCharColor*(x, y: INTEGER; fg, bg: INTEGER);
BEGIN
  ASSERT((x >= 0) & (y >= 0) & (x < charsX) & (y < charsY), 111);
  chars[y, x].fg := fg;
  chars[y, x].bg := bg
END SetCharColor;

PROCEDURE PutChar*(x, y: INTEGER; ch: CHAR; fg, bg: INTEGER);
BEGIN
  ASSERT((x >= 0) & (y >= 0) & (x < charsX) & (y < charsY), 111);
  chars[y, x].ch := ch;
  chars[y, x].fg := fg;
  chars[y, x].bg := bg
END PutChar;

PROCEDURE PutString*(x, y: INTEGER; s: ARRAY OF CHAR; fg, bg, limit: INTEGER);
VAR i: INTEGER;
BEGIN
  IF limit = 0 THEN limit := charsX END; 
  i := 0;
  WHILE (i < SHORT(LEN(s))) & (s[i] # 0X) & (x < limit) DO
    PutChar(x, y, s[i], fg, bg);
    INC(i); INC(x)
  END
END PutString;

PROCEDURE CharFill*(x1, y1, x2, y2: INTEGER; ch: CHAR; fg, bg: INTEGER);
VAR x, y: INTEGER;
BEGIN
  FOR y := y1 TO y2 DO
    FOR x := x1 TO x2 DO
      PutChar(x, y, ch, fg, bg)
    END
  END
END CharFill;

PROCEDURE ClearScreen*;
VAR x, y: INTEGER;
BEGIN
  FOR y := 0 TO charsY - 1 DO
    FOR x := 0 TO charsX - 1 DO
      chars[y, x].ch := ' ';
      chars[y, x].fg := 15;
      chars[y, x].bg := 1
    END
  END
END ClearScreen;

PROCEDURE ScrollScreen(lines: INTEGER);
VAR x, y: INTEGER;
BEGIN
  FOR y := 0 TO charsY - 1 - lines DO
    FOR x := 0 TO charsX - 1 DO
      chars[y, x] := chars[y + lines, x]
    END
  END;
  CharFill(0, charsY - lines, charsX - 1, charsY - 1, ' ', 7, 0) (*!FIXME colors*)
END ScrollScreen;

PROCEDURE Ln*;
BEGIN
  cursorX := 0;
  IF cursorY = charsY - 1 THEN ScrollScreen(1)
  ELSE INC(cursorY) END
END Ln;

PROCEDURE Backspace*;
BEGIN
  IF cursorX = 0 THEN
    cursorX := charsX - 1;
    DEC(cursorY)
  ELSE DEC(cursorX) END;
  chars[cursorY, cursorX].ch := ' '
END Backspace;

PROCEDURE Write*(ch: CHAR);
BEGIN
  IF ch = 0AX THEN Ln
  ELSIF ch # 0DX THEN
    PutChar(cursorX, cursorY, ch, 7, 0);
    IF cursorX = charsX - 1 THEN Ln
    ELSE INC(cursorX) END
  END;
END Write;

PROCEDURE WriteString*(s: ARRAY OF CHAR);
VAR i: INTEGER;
BEGIN
  i := 0;
  WHILE s[i] # 0X DO Write(s[i]); INC(i) END
END WriteString;

PROCEDURE LoadMedia(): BOOLEAN;
VAR success: BOOLEAN;
BEGIN
  font := G.LoadFont('data/images/font.bmp', charW, charH);
  success := font # NIL;
  RETURN success
END LoadMedia;

PROCEDURE Init*(fullscreen: BOOLEAN): BOOLEAN;
VAR success: BOOLEAN; options: SET;
BEGIN
  success := FALSE;
  options := {G.buffered, G.initMouse, G.spread};
  IF fullscreen THEN INCL(options, G.fullscreen) END;
  G.Settings(640, 400, options);
  G.SetSizeStep(charW, charH);
  screen := G.Init();
  IF screen # NIL THEN
    G.SetWindowTitle('FreeOberon');
    G.ShowMouse(FALSE);
    charsX := screen.w DIV charW;
    charsY := screen.h DIV charH;
    scrH := charsY - 4;
    IF LoadMedia() THEN
      success := TRUE;
      insertCursor := FALSE;
      cursorShown := TRUE;
      cursorX := 1;  cursorY := 2;
      mouseX := 0;  mouseY := 0;
      ResizeScreen
    END
  END;
  RETURN success
END Init;

END Terminal.
