MODULE Widgets;

IMPORT G := Graph, T := Terminal;

TYPE
  Widget* = POINTER TO WidgetDesc;
  WidgetDesc* = RECORD
    x*, y*, w*, h*: INTEGER
  END;

  (* Multiline text editing window *)
  EditingWindow* = POINTER TO EditingWindowDesc;
  EditingWindowDesc* = RECORD(WidgetDesc)
    (* !TODO *)
  END;

  (* Single line editing text box *)
  Edit* = POINTER TO EditDesc;
  EditDesc* = RECORD(WidgetDesc)
    s*: ARRAY 1024 OF CHAR;
    len*: INTEGER; (* Actual amount characters before 0X *)
    pos*: INTEGER
  END;

  (* Single line editing text box *)
  InfoBox* = POINTER TO InfoBoxDesc;
  InfoBoxDesc* = RECORD(WidgetDesc) END;

PROCEDURE InsertChar(ch: CHAR; pos: INTEGER;
  VAR s: ARRAY OF CHAR; VAR len: INTEGER);
VAR i: INTEGER;
BEGIN
  IF len + 1 < LEN(s) THEN
    FOR i := len TO pos + 1 BY -1 DO s[i] := s[i - 1] END;
    s[pos] := ch; INC(len); s[len] := 0X
  END
END InsertChar;

PROCEDURE DeleteChar(VAR s: ARRAY OF CHAR;
  pos: INTEGER; VAR len: INTEGER);
VAR i: INTEGER;
BEGIN
  FOR i := pos TO len - 1 DO s[i] := s[i + 1] END;
  DEC(len); s[len] := 0X
END DeleteChar;

(* Widget *)
PROCEDURE InitWidget*(wg: Widget; x, y, w, h: INTEGER);
BEGIN
  wg.x := x; wg.y := y; wg.w := w; wg.h := h
END InitWidget;

PROCEDURE (wg: Widget) HandleKeyDown*(key: INTEGER);
BEGIN
END HandleKeyDown;

PROCEDURE (wg: Widget) Draw*;
BEGIN
  T.CharFill(wg.x, wg.y, wg.x + wg.w - 1,
    wg.y + wg.h - 1, '?', 1, 2)
END Draw;

(* Editing Window *)
PROCEDURE InitEditingWindow*(e: EditingWindow; x, y, w, h: INTEGER);
BEGIN
  InitWidget(e, x, y, w, h)
END InitEditingWindow;

PROCEDURE NewEditingWindow*(x, y, w, h: INTEGER): EditingWindow;
VAR e: EditingWindow;
BEGIN
  NEW(e); InitEditingWindow(e, x, y, w, h); RETURN e
END NewEditingWindow;

(* Edit *)
PROCEDURE InitEdit*(e: Edit; x, y, w, h: INTEGER);
BEGIN
  InitWidget(e, x, y, w, h);
  e.s[0] := 0X; e.len := 0; e.pos := 0
END InitEdit;

PROCEDURE NewEdit*(x, y, w, h: INTEGER): Edit;
VAR e: Edit;
BEGIN
  NEW(e); InitEdit(e, x, y, w, h); RETURN e
END NewEdit;

PROCEDURE (e: Edit) HandleKeyDown*(key: INTEGER);
VAR code: INTEGER; ch: CHAR;
BEGIN
  code := G.GetScancode(key);
  CASE code OF
    G.kLeft : IF e.pos > 0 THEN DEC(e.pos) END
  | G.kRight: IF e.pos < e.len THEN INC(e.pos) END
  | G.kBackspace:
    IF e.pos > 0 THEN DeleteChar(e.s, e.pos - 1, e.len); DEC(e.pos) END
  | G.kDel: IF e.pos < e.len THEN DeleteChar(e.s, e.pos, e.len) END
  | G.kHome: e.pos := 0
  | G.kEnd: e.pos := e.len
  ELSE
    IF G.IsPrintable(key) THEN
      ch := G.GetChar(key);
      IF ch # 0X THEN
        InsertChar(ch, e.pos, e.s, e.len);
        IF e.pos < e.len THEN INC(e.pos) END
      END
    END
  END
END HandleKeyDown;

PROCEDURE (e: Edit) Draw*;
VAR i, x: INTEGER;
BEGIN
  T.PutChar(e.x, e.y, ' ', 0, 1);
  i := 0; x := e.x + 1;
  WHILE (e.s[i] # 0X) & (i < e.w) DO
    T.PutChar(x, e.y, e.s[i], 15, 1); INC(i); INC(x)
  END;
  WHILE i < e.w DO
    T.PutChar(x, e.y, ' ', 15, 1); INC(i); INC(x)
  END;
  T.GoToXY(e.x + e.pos + 1, e.y)
END Draw;

(* Info Box *)
PROCEDURE InitInfoBox*(e: InfoBox; x, y, w, h: INTEGER);
BEGIN
  InitWidget(e, x, y, w, h)
END InitInfoBox;

PROCEDURE NewInfoBox*(x, y, w, h: INTEGER): InfoBox;
VAR e: InfoBox;
BEGIN
  NEW(e); InitInfoBox(e, x, y, w, h); RETURN e
END NewInfoBox;

PROCEDURE (e: InfoBox) Draw*;
VAR i: INTEGER;
BEGIN
  FOR i := e.x + 1 TO e.x + e.w - 3 DO
    T.PutChar(i, e.y, 0CDX, 15, 3); (* = *)
    T.PutChar(i, e.y + e.h - 2, 0CDX, 15, 3);
    T.SetCharColor(i + 1, e.y + e.h - 1, 8, 0) (* Bottom shadow *)
  END;
  FOR i := e.y + 1 TO e.y + e.h - 2 DO
    T.PutChar(e.x, i, 0BAX, 15, 3); (* || *)
    T.PutChar(e.x + e.w - 3, i, 0BAX, 15, 3);
    T.SetCharColor(e.x + e.w - 2, i, 8, 0); (* Right shadow *)
    T.SetCharColor(e.x + e.w - 1, i, 8, 0)
  END;
  T.SetCharColor(e.x + e.w - 1, e.y + e.h - 1, 8, 0); (* Bottom-right shadow *)
  T.PutChar(e.x, e.y, 0C9X, 15, 3); (* Corners *)
  T.PutChar(e.x + e.w - 3, e.y, 0BBX, 15, 3);
  T.PutChar(e.x, e.y + e.h - 2, 0C8X, 15, 3);
  T.PutChar(e.x + e.w - 3, e.y + e.h - 2, 0BCX, 15, 3);
  T.CharFill(e.x + 1, e.y + 1, e.x + e.w - 4, e.y + e.h - 3, ' ', 0, 3)

END Draw;

END Widgets.
